"""Checks performed in other places, throwing errors when unsatisfied. """
from typing import TypeVar, Type, Optional

T = TypeVar('T')


def check_is_instance(value: T, type_: Type[T], message: Optional[str] = None) -> T:
    message = message or f'Got "{value}" when expecting a "{type_}" instance.'
    if not isinstance(value, type_):
        raise TypeError(message)
    return value



"""Descriptors of types used by Serious.

Descriptors are simplifying work with types, enriching them with more contextual information.
This allows to make decisions, like picking a serializer, easier.

They unwrap the generic aliases, get generic parameters from parent classes, simplify optional,
dataclass checks and more.

The data is carried by `TypeDescriptor`s which are created by a call to `serious.descriptors.describe(cls)`.
"""
from __future__ import annotations

__all__ = ['TypeDescriptor', 'describe', 'DescTypes', 'scan_types']

import datetime
import decimal
import uuid
from dataclasses import dataclass, fields, is_dataclass
from types import UnionType, NoneType
from typing import Type, Any, TypeVar, get_type_hints, Dict, Mapping, List, Union, Iterable, Optional, cast

from .types import FrozenDict, FrozenList

T = TypeVar('T')

GenericParams = Mapping[Any, 'TypeDescriptor']


@dataclass(frozen=True)
class TypeDescriptor:
    """A descriptor of a type unwrapping the aliases, optionals, separating generic parameters,
    extracting the parameters from broader context, etc.

    Type descriptors are mostly used for mapping serializers to particular objects.

    A proper way of creating a `TypeDescriptor` is using the `serious.descriptors.describe(cls)` factory.
    """
    _cls: Type
    parameters: FrozenDict[Any, TypeDescriptor]
    is_optional: bool = False
    is_dataclass: bool = False
    is_typed_dict: bool = False

    @property
    def is_sqlalchemy_model(self):
        from serious.serialization.field_serializers import SQLALCHEMY_INTEGRATION_ENABLED

        if not SQLALCHEMY_INTEGRATION_ENABLED:
            return False

        from sqlalchemy.orm import DeclarativeMeta

        return isinstance(self.cls, DeclarativeMeta)

    @property
    def cls(self):  # Python fails when providing cls as a keyword parameter to dataclasses
        return self._cls

    @property
    def fields(self) -> Mapping[str, TypeDescriptor]:
        """A mapping of all dataclass or typed dict field names to their corresponding Type Descriptors.

        An empty mapping is returned if the object is not a dataclass."""
        if self.is_dataclass:
            types = get_type_hints(self.cls)  # type: Dict[str, Type]
            descriptors = {name: self.describe(type_) for name, type_ in types.items()}
            return {f.name: descriptors[f.name] for f in fields(self.cls)}
        if self.is_typed_dict:
            types = get_type_hints(self.cls)
            descriptors = {name: self.describe(type_) for name, type_ in types.items()}
            return {key: descriptors[key] for key in self.cls.__annotations__}
        if self.is_sqlalchemy_model:
            _fields_names = [p.key for p in self._cls.__mapper__.attrs]
            mapped_types = get_type_hints(self.cls)
            descriptors = {
                name: self.describe(self._sqlalchemy_mapped_type(type_))
                for name, type_ in mapped_types.items()
            }
            for f in _fields_names:
                if f not in descriptors:
                    type_ = self.cls.__mapper__.columns[f].type
                    descriptors[f] = self.describe(_get_sqlalchemy_builtin_type(type_))
            return {f: descriptors[f] for f in _fields_names}
        return {}

    def _sqlalchemy_mapped_type(self, type_) -> Type:
        from sqlalchemy.orm import Mapped

        if issubclass(getattr(type_, "__origin__", NoneType), Mapped):
            return type_.__args__[0]  #type: ignore

        return type_

    def describe(self, type_: Type) -> TypeDescriptor:
        return describe(type_, self.parameters)

    def __repr__(self):
        return f"<TypeDescriptor {str(self)}>"

    def __str__(self):
        results = f""
        if self.is_optional:
            results += "Optional["
        results += getattr(self.cls, "__name__", str(self.cls))
        if self.parameters:
            results += (
                "[" + ", ".join(f"{k}={v}" for k, v in self.parameters.items()) + "]"
            )
        if self.is_optional:
            results += "]"
        return results

def describe(type_: Type, generic_params: Optional[GenericParams] = None) -> TypeDescriptor:
    """Creates a TypeDescriptor for the provided type.

    Optionally generic params can be designated as a mapping of TypeVar to parameter Type or indexes in Dict/List/etc.
    """
    generic_params = generic_params if generic_params is not None else {}
    param = generic_params.get(type_, None)
    if param is not None:
        return param
    return _describe_generic(type_, generic_params)


def _get_sqlalchemy_builtin_type(column_type):
    from sqlalchemy.sql.sqltypes import (
        String, Enum, Text, Unicode, UnicodeText, VARCHAR, NVARCHAR, CHAR, NCHAR, NullType,
        Integer, SmallInteger, BigInteger, Float, Double, REAL, NUMERIC, DECIMAL, Numeric,
        Boolean, DateTime, TIMESTAMP, DATETIME, Date, Time, Interval, LargeBinary, JSON, ARRAY, PickleType, UUID,
    )

    if isinstance(column_type, (String, Enum, Text, Unicode, UnicodeText, VARCHAR, NVARCHAR, CHAR, NCHAR)):
        if isinstance(column_type, Enum):
            return column_type.enum_class  # Return the specific enum class
        return str
    elif isinstance(column_type, (Integer, SmallInteger, BigInteger)):
        return int
    elif isinstance(column_type, (Float, Double, REAL, NUMERIC, DECIMAL)):
        return float
    elif isinstance(column_type, Numeric):
        return decimal.Decimal
    elif isinstance(column_type, Boolean):
        return bool
    elif isinstance(column_type, (DateTime, TIMESTAMP, DATETIME)):
        return datetime.datetime
    elif isinstance(column_type, Date):
        return datetime.date
    elif isinstance(column_type, Time):
        return datetime.time
    elif isinstance(column_type, Interval):
        return datetime.timedelta
    elif isinstance(column_type, LargeBinary):
        return bytes
    elif isinstance(column_type, JSON):
        return dict
    elif isinstance(column_type, ARRAY):
        return list
    elif isinstance(column_type, PickleType):
        raise NotImplementedError("PickleType is not supported")
    elif isinstance(column_type, UUID):
        return uuid.UUID
    elif isinstance(column_type, NullType):
        return type(None)
    else:
        raise NotImplementedError(f"Type {column_type} is not supported")

_any_type_desc = TypeDescriptor(Any, FrozenDict())  # type: ignore
_generic_params: Dict[Type, Dict[int, TypeDescriptor]] = {
    list: {0: _any_type_desc},
    set: {0: _any_type_desc},
    frozenset: {0: _any_type_desc},
    tuple: {0: _any_type_desc, 1: TypeDescriptor(Ellipsis, FrozenDict())},  # type: ignore
    dict: {0: _any_type_desc, 1: _any_type_desc},
    FrozenDict: {0: _any_type_desc, 1: _any_type_desc},
}


def _get_default_generic_params(cls: Type, params: GenericParams) -> GenericParams:
    """Returns mapping of default generic params for the provided cls.

    **Examples**:
     - `dict` -> `{0: <TypeDescriptor cls=Any>, 1: <TypeDescriptor cls=Any>}`;
     - `list` -> `{0: <TypeDescriptor cls=Any>}`;
     - `tuple` -> `{0: <TypeDescriptor cls=Any>, 1: <TypeDescriptor cls=Ellipses>}`.
    """
    for generic, default_params in _generic_params.items():
        if issubclass(cls, generic):
            return default_params
    return params


def _describe_generic(cls: Type, generic_params: GenericParams) -> TypeDescriptor:
    """Creates a TypeDescriptor for Python _GenericAlias, unwrapping it to its origin/

    **Examples**:
     - `Tuple[str]` -> `<TypeDescriptor cls=tuple params={0: <TypeDescriptor cls=str>}>`;
     - `Optional[int]` -> `<TypeDescriptor cls=int is_optional=True>`.
    """
    params: GenericParams = {}
    is_optional = _is_optional(cls)
    if is_optional:
        _args = set(cls.__args__)
        _args.remove(type(None))
        cls = cast(Type, Union[tuple(_args)])

    try:
        is_typed_dict = issubclass(cls, dict) and bool(getattr(cls, '__annotations__', None))
    except TypeError:
        is_typed_dict = False

    if hasattr(cls, '__orig_bases__') and is_dataclass(cls):
        _params: Dict[Any, TypeDescriptor] = {}
        for item in (_describe_generic(base, generic_params).parameters for base in getattr(cls, '__orig_bases__', [])):
            _params.update(item)

        return TypeDescriptor(
            cls,
            parameters=FrozenDict(_params),
            is_optional=is_optional,
            is_dataclass=True,
            is_typed_dict=False,
        )

    if hasattr(cls, '__origin__'):
        origin = cls.__origin__
        return _describe_parametrized(cls, generic_params, is_optional, is_typed_dict, origin)

    if isinstance(cls, UnionType):
        return _describe_parametrized(cls, generic_params, is_optional, is_typed_dict, Union)

    if isinstance(cls, type) and len(params) == 0 and not is_typed_dict:
        params = _get_default_generic_params(cls, params)
    return TypeDescriptor(
        cls,
        parameters=FrozenDict(params),
        is_optional=is_optional,
        is_dataclass=is_dataclass(cls),
        is_typed_dict=is_typed_dict,
    )


def _describe_parametrized(cls, generic_params, is_optional, is_typed_dict, origin):
    origin_is_dc = is_dataclass(origin)
    if origin_is_dc:
        params = _collect_type_vars(cls, generic_params)
    else:
        describe_ = lambda arg: describe(Any if type(arg) is TypeVar else arg, generic_params)
        params = dict(enumerate(map(describe_, getattr(cls, '__args__', []))))
    if isinstance(origin, type) and len(params) == 0:
        params = _get_default_generic_params(origin, params)
    descriptor = TypeDescriptor(
        origin,
        parameters=FrozenDict(params),
        is_optional=is_optional,
        is_dataclass=origin_is_dc,
        is_typed_dict=is_typed_dict,
    )
    return descriptor


def _collect_type_vars(alias: Any, generic_params: GenericParams) -> GenericParams:
    return dict(zip(alias.__origin__.__parameters__,
                    (describe(arg, generic_params) for arg in alias.__args__)))


class DescTypes:
    types: FrozenList[Type]

    def __init__(self, types: Iterable[Type]):
        super().__setattr__('types', FrozenList(types))

    @classmethod
    def scan(cls, desc: TypeDescriptor, *, known: List[TypeDescriptor]) -> 'DescTypes':
        if desc in known:
            return _empty_desc_types
        known.append(desc)
        dts = []  # type: List[DescTypes]
        for param in desc.parameters.values():
            dts.append(cls.scan(param, known=known))
        for child_desc in desc.fields.values():
            dts.append(cls.scan(child_desc, known=known))
        types = [type_ for dt in dts for type_ in dt.types]
        types.append(desc.cls)
        return cls(types)

    def __setattr__(self, key, value):
        raise AttributeError('Attempt to modify an immutable object')

    def __contains__(self, item):
        return item in self.types


_empty_desc_types = DescTypes({})


def scan_types(desc: TypeDescriptor) -> DescTypes:
    """Create a `DescTypes` object for the provided descriptor.

    `DescTypes` allow checks of the descriptor tree."""
    return DescTypes.scan(desc, known=[])


def _is_optional(cls: Type) -> bool:
    """Returns True if the provided type is `Optional`."""
    return (getattr(cls, '__origin__', None) == Union or isinstance(cls, UnionType)) \
        and len(cls.__args__) > 1 \
        and type(None) in set(cls.__args__)



"""
Test utilities for use with pytest.

Right now contains single check -- `assert_symmetric`.
With it’s possible to check that a dataclass stays the same after dumping and loading the dumped instance.
"""

__all__ = ['assert_symmetric']

from typing import overload

from .dict import DictModel
from .json import JsonModel
from .utils import Dataclass


@overload
def assert_symmetric(serializer: DictModel, value: Dataclass):
    pass


@overload
def assert_symmetric(serializer: JsonModel, value: Dataclass):
    pass


def assert_symmetric(serializer, value):
    """Asserts that dumping an instance of dataclass via model and loading it will result in an equal object."""
    assert serializer.load(serializer.dump(value)) == value, f'load/dump are not symmetric in {serializer}.'



"""Errors raised by serious.

Errors are divided into 3 main groups:

1. Validation errors -- raised by serious or library users if an objects fails validation.
Contains the stack pointing to an object which raised the error.
2. Serialization errors -- wrap the exceptions which occur during serialization.
They also contain the stack pointing to a specific place which raised the error.
Can be one of `LoadError` or `DumpError`.
3. Model errors -- raised when traversing the dataclass and building the model.
When no serializer found for field, etc.
"""
from __future__ import annotations

__all__ = [
    'SerializationError',
    'LoadError',
    'DumpError',
    'UnexpectedItem',
    'MissingField',
    'ModelError',
    'FieldMissingSerializer',
    'ModelContainsAny',
    'MutableTypesInModel',
    'ValidationError',
]

from typing import Type, Mapping, Collection, TYPE_CHECKING, Iterable

from .utils import class_path, Dataclass

if TYPE_CHECKING:
    from .serialization.context import SerializationStep
    from .descriptors import TypeDescriptor


class SerializationError(Exception):
    """Base for any non-validation error during load or dump."""

    def __init__(self, cls: Type, serializer_stack: Collection[SerializationStep]):
        self.cls = cls
        self._path = self.__parse_stack(serializer_stack)
        super().__init__(self.message)

    @staticmethod
    def __parse_stack(serializer_stack: Collection[SerializationStep]) -> str:
        if len(serializer_stack) == 0:
            return ''
        return ''.join(serializer_stack)[1:]

    @property
    def message(self):
        return f'Error during serialization of "{self.cls}"'

    def __str__(self):
        return self.message


class LoadError(SerializationError):
    """Non-validation error during construction of a dataclass instance from external data.

    This will wrap any error during load process.
    Either invalid data is supplied, or the serializers are handling data incorrectly.
    """

    def __init__(self, cls: Type, serializer_stack: Collection[SerializationStep], data: Mapping):
        self._data = data
        super().__init__(cls, serializer_stack)

    @property
    def message(self):
        return f'Failed to load "{self._path}" of {class_path(self.cls)}: {self.__cause__}. Data: {self._data}'


class DumpError(SerializationError):
    """Non-validation error during encoding of a dataclass instance.

    This error will wrap any error during dump and signifies that
    an invalid dataclass object got into serializers hands or serializers behave incorrectly.
    """

    def __init__(self, obj: Dataclass, serializer_stack: Collection[SerializationStep]):
        self._object = obj
        super().__init__(type(obj), serializer_stack)

    @property
    def message(self):
        if self.__cause__:
            return f"Failed to dump \"{self._path}\": {self.__cause__}. Object: {self._object}"
        return f'Failed to dump "{self._path}". Object: {self._object}'


class UnexpectedItem(LoadError):
    def __init__(self, cls: Type[Dataclass], data, fields: Collection[str]):
        self._fields = fields
        super().__init__(cls, [], data)

    @property
    def message(self):
        if len(self._fields) == 1:
            field = next(iter(self._fields))
            return f'Unexpected field "{field}" in loaded {class_path(self.cls)}'
        else:
            return f'Unexpected fields {self._fields} in loaded {class_path(self.cls)}'


class MissingField(LoadError):

    def __init__(self, cls: Type[Dataclass], data, fields: Collection[str]):
        self._fields = fields
        super().__init__(cls, [], data)

    @property
    def message(self):
        if len(self._fields) == 1:
            field = next(iter(self._fields))
            return f'Missing field "{field}" in loaded {class_path(self.cls)}'
        else:
            return f'Missing fields {self._fields} in loaded {class_path(self.cls)}'


class ModelError(Exception):
    """An error that can occur when traversing class structure.

    Library and the user can place restrictions on what the model can be.
    This would include containing Any, model mutability, etc.
    Instances of this error will be raised during model construction if such restrictions are unsatisfied.
    """

    def __init__(self, cls: Type):
        self.cls = cls
        super().__init__(self.message)

    @property
    def message(self):
        return f'Model error in class "{self.cls}ю"'


class FieldMissingSerializer(ModelError):

    def __init__(self, cls: Type, desc: TypeDescriptor):
        self.desc = desc
        super().__init__(cls)

    @property
    def message(self):
        return (f'{class_path(self.cls)} contains unserializable member: {self.desc}. '
                f'Create a serializer fitting the descriptor and pass it to the model ``serializers``.')


class ModelContainsAny(ModelError):

    @property
    def message(self):
        return (f'{class_path(self.cls)} contains fields annotated as Any or missing type annotation. '
                f'Provide a type annotation or pass `allow_any=True` to the serializer. '
                f'This may also be an ambiguous ``Generic`` definitions like `x: list`, `x: List` '
                f'which are resolved as `List[Any]`. ')


class MutableTypesInModel(ModelError):

    def __init__(self, cls: Type, mutable_types: Iterable[Type]):
        self.mutable_types = mutable_types
        super().__init__(cls)

    @property
    def message(self):
        return (f'{class_path(self.cls)} is has mutable members: {self.mutable_types}.'
                f'If there are immutable types pass them to model as `ensure_frozen=[Type1, Type2]`.'
                f'Replace mutable types with frozen ones. '
                f'Set @dataclass(frozen=True). \n'
                f'Alternatively, allow mutable fields by passing `ensure_frozen=False` to model. ')


class ValidationError(Exception):
    """An error manifesting an invalid object state.

    Raised when calling `validate(obj)` and internally by serious during load/dump.

    Contains the stack pointing to an exact point where validation failed in data structure.
    """

    def __init__(self, message='Failed validation'):
        super().__init__(message)



"""Checks performed in other places, throwing errors when unsatisfied."""

import collections
from typing import Any, Type

from serious.json.errors import UnexpectedJson


def check_that_loading_an_object(data: Any, cls: Type):
    """Checks data is a Mapping. If not raises an `serious.json.errors.UnexpectedJson` with a helpful error message."""

    if not isinstance(data, collections.abc.Mapping):
        if isinstance(data, collections.abc.Collection):
            raise UnexpectedJson(f'Expecting a single object in JSON, got a collection instead. '
                                 f'Use #load_all(cls) instead of #load(cls) '
                                 f'to decode an array of {cls} dataclasses.')
        raise UnexpectedJson(f'Expecting a single {cls} object encoded in JSON.')


def check_that_loading_a_list(data: Any, cls: Type):
    """Checks data is a Collection but not a Mapping.
    If not raises an `serious.json.errors.UnexpectedJson` with a helpful error message.
    """

    if not isinstance(data, collections.abc.Collection):
        raise UnexpectedJson(f'Expecting an array of {cls} objects encoded in JSON.')
    if isinstance(data, collections.abc.Mapping):
        raise UnexpectedJson(f'Expecting an array of objects encoded in JSON, got a mapping instead.'
                             f'Use #load(cls) instead of #load_all(cls) '
                             f'to decode a single {cls} dataclasses.')



"""Check that whole dataclass structure is immutable, i.e. it’s objects cannot be changed. """
__all__ = ["check_immutable"]

from dataclasses import is_dataclass
from datetime import datetime, date, time
from decimal import Decimal
from enum import Enum
from typing import Iterable, Type, List, Any, Union
from uuid import UUID

from serious.descriptors import DescTypes, TypeDescriptor
from serious.errors import MutableTypesInModel
from serious.types import FrozenList, Email, Timestamp, FrozenDict

_IMMUTABLE_TYPES = {
    str, int, float, bool,
    bytes, tuple, frozenset, FrozenList, FrozenDict,
    Decimal, UUID, datetime, date, time,
    Email, Timestamp,
    Ellipsis,
}


def check_immutable(desc: TypeDescriptor, all_types: DescTypes, ensure_frozen: Union[bool, Iterable[Type]]):
    user_frozen = ensure_frozen if isinstance(ensure_frozen, Iterable) else {}
    mutable_types = extract_mutable(all_types, also_immutable=user_frozen)
    if len(mutable_types):
        raise MutableTypesInModel(desc.cls, mutable_types)


def extract_mutable(desc: DescTypes, also_immutable: Iterable[Type]) -> List[Type]:
    allowed_types = _IMMUTABLE_TYPES | set(also_immutable)
    maybe_dc = set(desc.types) - allowed_types
    restricted = [type_ for type_ in maybe_dc if not is_frozen_dc(type_) and not issubclass(type_, Enum)]
    return restricted


def is_frozen_dc(type_: Any) -> bool:
    return is_dataclass(type_) and type_.__dataclass_params__.frozen



"""Serializable types provided by serious.

`serious.types` module lists a number of types useful for as part of serious modules:
- Timestamp — an alternative to datetime, that is serialized to a float ms value
- Email — a Tiny Type made out of string, which is checked to match the format
- FrozenList and FrozenDict — immutable alternatives for standard collections
"""

from __future__ import annotations

__all__ = ['Timestamp', 'Email', 'FrozenList', 'FrozenDict']

import re
from datetime import datetime, timezone
from typing import TypeVar, Generic, overload, Optional, Mapping, Iterator

from .errors import ValidationError

KT = TypeVar('KT')  # Key type.
VT = TypeVar('VT')  # Value type.


class FrozenDict(Mapping[KT, VT]):
    """
    A dictionary which cannot be changed a la frozenset.

    Does not check for immutability of its members.
    """

    def __init__(self, mapping: Optional[Mapping[KT, VT]] = None, **kwargs: Mapping[KT, VT]) -> None:
        if mapping is not None:
            self.__internal_mapping__ = dict(mapping)
        else:
            self.__internal_mapping__ = dict(**kwargs)

    def __getitem__(self, __k: KT) -> VT:
        return self.__internal_mapping__[__k]

    def __len__(self) -> int:
        return len(self.__internal_mapping__)

    def __iter__(self) -> Iterator[KT]:
        return iter(self.__internal_mapping__)

    def __hash__(self):
        hash_ = 0
        for key, value in self.items():
            hash_ ^= hash((key, value))
        return hash_

    def __or__(self, other: Mapping) -> FrozenDict:
        if hasattr(other, 'items'):
            result = dict(self)
            result.update(other)
            return FrozenDict(result)
        raise TypeError(f'Cannot merge {type(self)} with {type(other)}')

    def __repr__(self):
        return f'FrozenDict({self.__internal_mapping__})'


class FrozenList(tuple, Generic[VT]):
    """A list that cannot be changed. `FrozenList[VT]` is equal to saying `Tuple[VT, ...]`."""
    pass


class Timestamp:
    """An immutable(frozen) UTC timestamp.

    Constructor accepts a single arguments. One of:
    - datetime object
    - ISO 8601 formatted string
    - number of seconds since UNIX epoch in int or float
    - a Timestamp object to create its copy

    Supports comparison with another Timestamp objects.
    """

    value: float

    @overload
    def __init__(self, dt: datetime):
        """Creates a timestamp matching the value of a datetime instance."""
        pass

    @overload
    def __init__(self, iso_string: str):
        """Creates a timestamp from a provided ISO 8601 formatted string."""
        pass

    @overload
    def __init__(self, timestamp: int):
        """Creates a timestamp with provided number of seconds since UNIX epoch (January 1st, 1970)."""
        pass

    @overload
    def __init__(self, timestamp: float):
        """Creates a timestamp with provided number of seconds since UNIX epoch (January 1st, 1970)."""
        pass

    @overload
    def __init__(self, timestamp: 'Timestamp'):
        """Creates a copy of a timestamp."""
        pass

    def __init__(self, value):
        if isinstance(value, Timestamp):
            super().__setattr__('value', value.value)
        elif isinstance(value, datetime):
            super().__setattr__('value', self._datetime_value(value))
        elif isinstance(value, str):
            dt = datetime.fromisoformat(value)
            super().__setattr__('value', self._datetime_value(dt))
        elif isinstance(value, int):
            super().__setattr__('value', float(value))
        elif isinstance(value, float):
            super().__setattr__('value', value)
        else:
            raise ValueError(f'Timestamp cannot be created from "{value}"')

    @staticmethod
    def _datetime_value(value):
        offset = value.utcoffset()
        if offset:
            value = value - offset
        return value.timestamp()

    def as_datetime(self) -> datetime:
        """Constructs a datetime object with the timestamp value."""
        return datetime.fromtimestamp(self.value, tz=timezone.utc)

    def as_iso(self):
        """ISO 8601 representation of a timestamp. Same as str(timestamp(value))"""
        return datetime.fromtimestamp(self.value, tz=timezone.utc).isoformat()

    def __setattr__(self, name, value):
        """Restricts mutation of object value"""
        raise AttributeError('Cannot change timestamp. Timestamp objects are immutable.')

    def __eq__(self, other: object):
        """Overrides the default implementation"""
        if not isinstance(other, Timestamp):
            raise NotImplementedError('Cannot compare Timestamp to anything other than Timestamp.')
        return self.value == other.value

    def __lt__(self, other: 'Timestamp'):
        if not isinstance(other, Timestamp):
            raise NotImplementedError('Cannot compare Timestamp to anything other than Timestamp.')
        return self.value < other.value

    def __le__(self, other: 'Timestamp'):
        if not isinstance(other, Timestamp):
            raise NotImplementedError('Cannot compare Timestamp to anything other than Timestamp.')
        return self.value <= other.value

    def __ge__(self, other: 'Timestamp'):
        if not isinstance(other, Timestamp):
            raise NotImplementedError('Cannot compare Timestamp to anything other than Timestamp.')
        return self.value >= other.value

    def __gt__(self, other: 'Timestamp'):
        if not isinstance(other, Timestamp):
            raise NotImplementedError('Cannot compare Timestamp to anything other than Timestamp.')
        return self.value > other.value

    def __str__(self):
        """ISO 8601 representation of a timestamp."""
        return self.as_iso()

    def __repr__(self):
        """An unambiguous representation of an object."""
        iso_str = self.as_iso()
        return f'<{self.__class__.__name__} {iso_str} ({self.value})>'


_email_regex = re.compile(r'^\w(\.|\w|-)*\+?(\.|\w|-)*@\w(\.|\w|-)*(\.\w+)$',
                          re.IGNORECASE | re.UNICODE)


class Email(str):
    """A regular email address. Email parts can be accessed via properties: `<username>+<label>@<domain>`."""

    def __new__(cls, content: str):
        return super().__new__(cls, content.lower())  # type: ignore # __new__ is a staticmethod

    @property
    def username(self) -> str:
        return self.split('@')[0].split('+')[0]

    @property
    def label(self) -> Optional[str]:
        user_and_label = self.split('@')[0].split('+')
        if len(user_and_label) == 1:
            return None
        return user_and_label[1]

    @property
    def domain(self) -> str:
        return self.split('@')[1]

    def __validate__(self):
        if _email_regex.match(self) is None:
            raise ValidationError('Invalid email format')

    def __repr__(self):
        return f'<{self.__class__.__name__} {str(self)}>'



"""
Serious has a simple understanding of what validation should be: a `__validate__` instance method raising ValidationError.

First of all, this puts all of validation together instead spreading it all over the place simplifying understanding.

Secondly, explicit `raise ValidationError` hits can be seen in test coverage which reduces possibility of missing them
in tests.

Serious runs objects `__validate__` method if it’s present during object load by default.
This can be overridden by passing `validate_on_load` and `validate_on_dump` to model.


You can run validation yourself by calling `serious.validation.validate(obj)` whenever you need.
"""
__all__ = ['validate']

from typing import TypeVar

T = TypeVar('T')


def validate(obj: T) -> T:
    """Executes objects own validation.

    Serious defines this validation as an instance `__validate__` method raising `ValidationError`.
    It also does not return nothing or returns None.

        :Example:

        @dataclass
        class Note:
            title: str
            content: str

            def __validate__(self):
                if not len(self.content):
                    raise ValidationError('Note cannot be empty')

        valid = Note('', 'A test!')
        validate(valid)  # nothing happens

        invalid = Note('', '')
        validate(invalid)  # raises ValidationError(...)
    """
    if hasattr(obj, '__validate__'):
        if obj.__validate__.__module__.startswith("pydantic"):
            return obj.__validate__(obj)
        result = obj.__validate__()  # type: ignore # method presence checked above
        assert result is None, 'Validators should not return anything. Raise ValidationError instead'
    return obj



"""A module with `JsonModule` -- Serious model to transform between dataclasses and JSON strings."""
from __future__ import annotations

__all__ = ['JsonModel']

import json
from typing import Optional, TypeVar, Type, Generic, List, MutableMapping, Collection, Iterable, Any, Union

from serious.descriptors import describe
from serious.serialization import FieldSerializer, SeriousModel, field_serializers, KeyMapper
from serious.utils import class_path
from serious.json.utils import camel_to_snake, snake_to_camel
from .checks import check_that_loading_an_object, check_that_loading_a_list

T = TypeVar('T')


class JsonModel(Generic[T]):
    """A model converting dataclasses to JSON strings and back.


        :Example:

        from uuid import UUID
        from dataclasses import dataclass
        from serious import DictModel

        @dataclass
        class Robot:
            serial: UUID
            name: str

        >>> model = JsonModel(Robot)
        >>> model.load('{"serial": "f3179d05-30f6-43ba-b6cb-7556af09330b", "name": "Caliban"}')
        Robot(serial=UUID('f3179d05-30f6-43ba-b6cb-7556af09330b'), name='Caliban')
        >>> model.dump(Robot(UUID('00000000-0000-4000-0000-000002716057'), 'Bender'))
        '{"serial": "00000000-0000-4000-0000-000002716057", "name": "Bender"}'

    Check `__init__` parameters for a list of configuration options.

    `More on models in docs <https://serious.readthedocs.io/en/latest/models/>`_.
    """

    def __init__(
            self,
            cls: Type[T],
            serializers: Iterable[Type[FieldSerializer]] = field_serializers(),
            *,
            allow_any: bool = False,
            allow_missing: bool = False,
            allow_unexpected: bool = False,
            validate_on_load: bool = True,
            validate_on_dump: bool = False,
            ensure_frozen: Union[bool, Iterable[Type]] = False,
            camel_case: bool = True,
            indent: Optional[int] = None,
    ):
        """Initialize a JSON model.

        :param cls: the dataclass type to load/dump.
        :param serializers: field serializer classes in an order they will be tested for fitness for each field.
        :param allow_any: `False` to raise if the model contains fields annotated with `Any`
                (this includes generics like `List[Any]`, or simply `list`).
        :param allow_missing: `False` to raise during load if data is missing the optional fields.
        :param allow_unexpected: `False` to raise during load if data contains some unknown fields.
        :param validate_on_load: to call dataclass `__validate__` method after object construction.
        :param validate_on_load: to call object `__validate__` before dumping.
        :param ensure_frozen: `False` to skip check of model immutability; `True` will perform the check
                against built-in immutable types; a list of custom immutable types is added to built-ins.
        :param camel_case: `True` to transform dataclass "snake_case" to JSON "camelCase".
        :param indent: number of spaces JSON output will be indented by; `None` for most compact representation.
        """
        self.cls = cls
        self.descriptor = describe(cls)
        self.serious_model: SeriousModel = SeriousModel(
            self.descriptor,
            serializers,
            allow_any=allow_any,
            allow_missing=allow_missing,
            allow_unexpected=allow_unexpected,
            validate_on_load=validate_on_load,
            validate_on_dump=validate_on_dump,
            ensure_frozen=ensure_frozen,
            key_mapper=JsonKeyMapper() if camel_case else None,
        )
        self._dump_indentation = indent

    def load(self, json_: str) -> T:
        """Load a dataclass from a JSON string."""
        data: MutableMapping = self._load_from_str(json_)
        check_that_loading_an_object(data, self.cls)
        return self.serious_model.load(data)

    def load_many(self, json_: str) -> List[T]:
        """Load a list of dataclasses from a JSON string."""
        data: Collection = self._load_from_str(json_)
        check_that_loading_a_list(data, self.cls)
        return [self.serious_model.load(each) for each in data]

    def dump(self, o: T) -> str:
        """Dump a single dataclass to a JSON string."""
        as_dict = self.serious_model.dump(o)
        return self._dump_to_str(as_dict)

    def dump_many(self, items: Collection[T]) -> str:
        """Dump a list of dataclasses to a JSON string."""
        as_dicts = [self.serious_model.dump(o) for o in items]
        return self._dump_to_str(as_dicts)

    def _load_from_str(self, json_: str) -> Any:
        """Override to customize JSON loading behaviour."""
        return json.loads(json_)

    def _dump_to_str(self, dict_items: Any) -> str:
        """Override to customize JSON dumping behaviour."""
        return json.dumps(dict_items,
                          skipkeys=False,
                          ensure_ascii=False,
                          check_circular=True,
                          allow_nan=False,
                          indent=self._dump_indentation,
                          separators=None,
                          default=None,
                          sort_keys=False)

    def __repr__(self):
        path = class_path(type(self))
        if path == 'serious.json.model.JsonModel':
            path = 'serious.JsonModel'
        return f'<{path}[{class_path(self.cls)}] at {hex(id(self))}>'


class JsonKeyMapper(KeyMapper):

    def to_model(self, key: str) -> str:
        return camel_to_snake(key)

    def to_serialized(self, field: str) -> str:
        return snake_to_camel(field)



"""Minor utilities used throughout the project."""

__all__ = ['class_path', 'Dataclass']

from typing import Type, Any

Dataclass = Any  # a dataclass instance


def class_path(cls: Type) -> str:
    """Returns a fully qualified type name."""
    return f'{cls.__module__}.{cls.__qualname__}'
